#[macro_use]
//pub extern crate diesel;

pub mod dtos;
mod models;
mod schema;

use crate::dtos::DateTime2;
use crate::models::{DepartureRequest, Movement, ServingLine, StationHint, Stop};
use chrono::{Local, NaiveDate, NaiveDateTime};
use sqlx::postgres::PgQueryResult;
use sqlx::PgPool;
use std::option::Option;
use std::{error::Error, str::FromStr};
use uuid::Uuid;

pub async fn insert_document(data: dtos::Root, conn: &mut PgPool) -> Result<(), Box<dyn Error>> {
    let _empty = vec![];
    let id = Uuid::new_v5(
        &Uuid::NAMESPACE_DNS,
        serde_json::to_string(&data.parameters)?.as_bytes(),
    );
    let req = if let Ok(req) = insert_req(
        conn,
        DepartureRequest {
            id,
            stop_id: data.dm.itd_odv_assigned_stops.stop_id.parse().ok(),
            server_id: data
                .parameters
                .iter()
                .filter(|p| p.name == "serverID")
                .map(|p| p.clone().value)
                .next(),
            request_id: data
                .parameters
                .iter()
                .filter(|p| p.name == "requestID")
                .map(|p| p.clone().value)
                .next(),
            session_id: data
                .parameters
                .iter()
                .filter(|p| p.name == "sessionID")
                .map(|p| p.clone().value)
                .next(),
            calc_time: f64::from_str(
                &data
                    .parameters
                    .iter()
                    .filter(|p| p.name == "calcTime")
                    .map(|p| p.clone().value)
                    .next()
                    .unwrap(),
            )
            .ok(),
            server_time: data
                .parameters
                .iter()
                .filter(|p| p.name == "serverTime")
                .map(|p| p.clone().value)
                .next()
                .unwrap()
                .parse::<NaiveDateTime>()
                .ok(),
            log_request_id: data
                .parameters
                .iter()
                .filter(|p| p.name == "logRequestId")
                .map(|p| p.clone().value)
                .next(),
        },
    )
    .await
    {
        if let Some(req) = req {
            req
        } else {
            id
        }
    } else {
        id
    };

    insert_stop(
        conn,
        &Stop {
            id: data.dm.input.input.parse().ok(),
            usage: data.dm.points.point.usage,
            r#type: data.dm.points.point.type_field.clone(),
            name: data.dm.points.point.name,
            anytype: data.dm.points.point.any_type,
            sort: data.dm.points.point.sort.parse().ok(),
            quality: data.dm.points.point.quality.parse().ok(),
            best: data.dm.points.point.best.parse().ok(),
            object: data.dm.points.point.object,
            gid: data.dm.points.point.ref_field.gid,
            omc: data.dm.points.point.ref_field.omc.parse().ok(),
            place_id: data.dm.points.point.ref_field.place_id.parse().ok(),
            x: if let Some(x) = data.dm.itd_odv_assigned_stops.x {
                x.parse().ok()
            } else {
                None
            },
            y: if let Some(y) = data.dm.itd_odv_assigned_stops.y {
                y.parse().ok()
            } else {
                None
            },
            map_name: data.dm.itd_odv_assigned_stops.map_name,
            place: data.dm.itd_odv_assigned_stops.place,
            name_with_place: data.dm.itd_odv_assigned_stops.name_with_place,
            distance_time: data.dm.itd_odv_assigned_stops.distance_time.parse().ok(),
            is_transfer_stop: data.dm.itd_odv_assigned_stops.is_transfer_stop.parse().ok(),
            vm: data.dm.itd_odv_assigned_stops.vm.parse().ok(),
            last_modified: Local::now().naive_local(),
        },
    )
    .await
    .unwrap();

    let _ = insert_line(
        conn,
        if let Some(lines) = &data.serving_lines.lines.map(|line| {
            line.into_iter()
                .filter_map(|line| {
                    if let Some(line) = line {
                        Some(ServingLine {
                            id: Uuid::new_v5(
                                &Uuid::NAMESPACE_DNS,
                                serde_json::to_string(&line).unwrap().as_bytes(),
                            ),
                            name: line.mode.name,
                            number: line.mode.number.parse().ok(),
                            product: line.mode.product,
                            product_id: line.mode.product_id.parse().ok(),
                            r#type: line.mode.type_field.parse().ok(),
                            code: line.mode.code.parse().ok(),
                            destination: line.mode.destination,
                            dest_id: line.mode.dest_id.parse().ok(),
                            desc: line.mode.desc,
                            timetable_period: line.mode.timetable_period,
                            branch: line.mode.diva.branch,
                            stateless: line.mode.diva.stateless,
                            trip_code: line.mode.diva.trip_code.parse().ok(),
                            operator: line.mode.diva.operator,
                            op_public_code: line.mode.diva.op_public_code,
                            op_code: line.mode.diva.op_code,
                            v_f: NaiveDate::parse_from_str(&line.mode.diva.v_f, "%Y%m%d").ok(),
                            v_to: NaiveDate::parse_from_str(&line.mode.diva.v_to, "%Y%m%d").ok(),
                            line_display: line.mode.diva.line_display,
                            last_modified: if let Ok(date) = data
                                .parameters
                                .iter()
                                .filter(|p| p.name == "serverTime")
                                .map(|p| p.clone().value)
                                .next()
                                .unwrap()
                                .parse::<NaiveDateTime>()
                            {
                                date
                            } else {
                                Local::now().naive_local()
                            },
                        })
                    } else {
                        None
                    }
                })
                .collect::<Vec<_>>()
        }) {
            lines
        } else {
            &_empty
        },
    )
    .await
    .unwrap();

    if let Some(infos) = data.dm.points.point.infos {
        insert_hints(
            conn,
            &StationHint {
                info_link_url: infos.info.info_link_url,
                info_link_text: infos.info.info_link_text,
                info_text: infos.info.info_text.content,
            },
        )
        .await
        .unwrap();
    };

    if let Some(arrival_list) = data.arrival_list {
        let _ = insert_arrivals(
            conn,
            &arrival_list
                .into_iter()
                .map(|arrival| Movement {
                    departure_request_id: req,
                    stop_id: arrival.stop_id.parse().ok(),
                    x: arrival.x.parse().ok(),
                    y: arrival.y.parse().ok(),
                    map_name: arrival.map_name,
                    area: arrival.area,
                    platform: arrival.platform,
                    platform_name: arrival.platform_name,
                    countdown: i32::from_str(&arrival.countdown).ok(),
                    time: datetime_from_dto(&arrival.date_time),
                    real_time: if let Some(date_time) = arrival.real_date_time {
                        datetime_from_dto(&date_time)
                    } else {
                        None
                    },
                    key: i32::from_str(&arrival.serving_line.key)
                        .ok()
                        .unwrap_or_default(),
                    code: arrival.serving_line.code,
                    number: arrival.serving_line.number,
                    symbol: arrival.serving_line.symbol,
                    mot_type: arrival.serving_line.mot_type.parse().ok(),
                    mt_subcode: arrival.serving_line.mt_subcode.parse().ok(),
                    realtime: arrival.serving_line.realtime.parse().ok(),
                    direction: arrival.serving_line.direction,
                    direction_from: arrival.serving_line.direction_from,
                    name: arrival.serving_line.name,
                    delay: if let Some(delay) = arrival.serving_line.delay {
                        i32::from_str(&delay).ok()
                    } else {
                        None
                    },
                    dest_id: arrival.serving_line.dest_id,
                    stateless: arrival.serving_line.stateless,
                    line_display: arrival.serving_line.line_display,
                    operator_code: arrival.operator.clone().map(|op| op.code),
                    operator_name: arrival.operator.clone().map(|op| op.name),
                    operator_public_code: arrival.operator.map(|op| op.public_code),
                    attrs: if let Some(attrs) = arrival.attrs {
                        attrs
                    } else {
                        vec![]
                    },
                    r#type: false,
                })
                .collect::<Vec<_>>(),
        )
        .await
        .unwrap();
    }

    if let Some(arrival_list) = data.departure_list {
        let _ = insert_arrivals(
            conn,
            &arrival_list
                .into_iter()
                .map(|arrival| Movement {
                    departure_request_id: req,
                    stop_id: i32::from_str(&arrival.stop_id).ok(),
                    x: if let Some(x) = arrival.x {
                        x.parse().ok()
                    } else {
                        None
                    },
                    y: if let Some(y) = arrival.y {
                        y.parse().ok()
                    } else {
                        None
                    },
                    map_name: arrival.map_name,
                    area: arrival.area,
                    platform: arrival.platform,
                    platform_name: arrival.platform_name,
                    countdown: i32::from_str(&arrival.countdown).ok(),
                    time: datetime_from_dto(&arrival.date_time),
                    real_time: if let Some(date_time) = arrival.real_date_time {
                        datetime_from_dto(&date_time)
                    } else {
                        None
                    },
                    key: i32::from_str(&arrival.serving_line.key)
                        .ok()
                        .unwrap_or_default(),
                    code: arrival.serving_line.code,
                    number: arrival.serving_line.number,
                    symbol: arrival.serving_line.symbol,
                    mot_type: arrival.serving_line.mot_type.parse().ok(),
                    mt_subcode: arrival.serving_line.mt_subcode.parse().ok(),
                    realtime: arrival.serving_line.realtime.parse().ok(),
                    direction: arrival.serving_line.direction,
                    direction_from: arrival.serving_line.direction_from,
                    name: arrival.serving_line.name,
                    delay: if let Some(delay) = arrival.serving_line.delay {
                        i32::from_str(&delay).ok()
                    } else {
                        None
                    },
                    dest_id: arrival.serving_line.dest_id,
                    stateless: arrival.serving_line.stateless,
                    line_display: arrival.serving_line.line_display,
                    operator_code: arrival.operator.clone().map(|op| op.code),
                    operator_name: arrival.operator.clone().map(|op| op.name),
                    operator_public_code: arrival.operator.map(|op| op.public_code),
                    attrs: if let Some(attrs) = arrival.attrs {
                        attrs
                    } else {
                        vec![]
                    },
                    r#type: true,
                })
                .collect::<Vec<_>>(),
        )
        .await
        .unwrap();
    }
    Ok(())
}
/*
async fn setup(conn: &PgPool) -> Result<(), sqlx::Error> {
    let sql = include_str!("../scheme.sql");
    sqlx::query(sql).execute(conn).await?;
    Ok(())
}*/

async fn insert_req(conn: &PgPool, req: DepartureRequest) -> Result<Option<Uuid>, sqlx::Error> {
    sqlx::query_scalar!(
        "INSERT INTO public.departure_requests VALUES ($1, $2, $3, $4, $5, $6, $7, $8) \
         ON CONFLICT (id) DO NOTHING \
         RETURNING id",
        req.id,
        req.stop_id,
        req.server_id,
        req.request_id,
        req.session_id,
        req.calc_time,
        req.server_time,
        req.log_request_id
    )
    .fetch_optional(conn)
    .await
}

async fn insert_stop(conn: &PgPool, stop: &Stop) -> Result<Vec<i32>, sqlx::Error> {
    sqlx::query_scalar!(
        "INSERT INTO stops VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14,
         $15, $16, $17, $18, $19, $20, $21)
         ON CONFLICT (id) DO NOTHING
         RETURNING id",
        stop.id,
        &stop.usage,
        stop.r#type,
        stop.name,
        stop.anytype,
        stop.sort,
        stop.quality,
        stop.best,
        stop.object,
        stop.gid,
        stop.omc,
        stop.place_id,
        stop.x,
        stop.y,
        stop.map_name,
        stop.place,
        stop.name_with_place,
        stop.distance_time,
        stop.is_transfer_stop,
        stop.vm,
        stop.last_modified
    )
    .fetch_all(conn)
    .await
}

async fn insert_line(conn: &PgPool, line: &Vec<ServingLine>) -> Result<Vec<Uuid>, sqlx::Error> {
    sqlx::query_scalar!(
        "INSERT INTO serving_lines SELECT * FROM
         UNNEST($1::uuid[], $2::varchar[], $3::integer[], $4::varchar[], $5::integer[],
         $6::integer[], $7::integer[], $8::varchar[], $9::integer[], $10::text[], $11::varchar[],
         $12::varchar[], $13::varchar[], $14::integer[], $15::varchar[], $16::varchar[],
         $17::varchar[], $18::date[], $19::date[], $20::varchar[], $21::timestamp[])
         ON CONFLICT (id) DO NOTHING
         RETURNING id",
        &line.iter().map(|sv| sv.id).collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.name.clone()).collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.number).collect::<Vec<_>>() as &[Option<_>],
        &line.iter().map(|sv| sv.product.clone()).collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.product_id).collect::<Vec<_>>() as &[Option<_>],
        &line.iter().map(|sv| sv.r#type).collect::<Vec<_>>() as &[Option<_>],
        &line.iter().map(|sv| sv.code).collect::<Vec<_>>() as &[Option<_>],
        &line
            .iter()
            .map(|sv| sv.destination.clone())
            .collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.dest_id).collect::<Vec<_>>() as &[Option<_>],
        &line.iter().map(|sv| sv.desc.clone()).collect::<Vec<_>>(),
        &line
            .iter()
            .map(|sv| sv.timetable_period.clone())
            .collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.branch.clone()).collect::<Vec<_>>(),
        &line
            .iter()
            .map(|sv| sv.stateless.clone())
            .collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.trip_code).collect::<Vec<_>>() as &[Option<_>],
        &line
            .iter()
            .map(|sv| sv.operator.clone())
            .collect::<Vec<_>>(),
        &line
            .iter()
            .map(|sv| sv.op_public_code.clone())
            .collect::<Vec<_>>() as &[Option<_>],
        &line.iter().map(|sv| sv.op_code.clone()).collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.v_f).collect::<Vec<_>>() as &[Option<_>],
        &line.iter().map(|sv| sv.v_to).collect::<Vec<_>>() as &[Option<_>],
        &line
            .iter()
            .map(|sv| sv.line_display.clone())
            .collect::<Vec<_>>(),
        &line.iter().map(|sv| sv.last_modified).collect::<Vec<_>>(),
    )
    .fetch_all(conn)
    .await
}

/*
async fn insert_req_line(
    conn: &PgPool,
    req: &Vec<ServingLineDepatureRequest>,
) -> Result<Vec<i32>, sqlx::Error> {
    sqlx::query_scalar!(
    "INSERT INTO serving_line_departure_requests (departureRequestId, servingLineId) SELECT * FROM \
     UNNEST($1::uuid[], $2::uuid[]) ON CONFLICT (id) DO NOTHING returning id",
     &req.iter().map(|r| r.departure_request_id).collect::<Vec<_>>(),
     &req.iter().map(|r| r.id).collect::<Vec<_>>()).fetch_all(conn).await
}
*/

async fn insert_hints(conn: &PgPool, hints: &StationHint) -> Result<PgQueryResult, sqlx::Error> {
    sqlx::query!(
        "INSERT INTO station_hints VALUES($1, $2, $3)
         ON CONFLICT (infoLinkURL) DO NOTHING",
        hints.info_link_url,
        hints.info_link_text,
        hints.info_text
    )
    .execute(conn)
    .await
}

async fn insert_arrivals(
    conn: &PgPool,
    arrivals: &Vec<Movement>,
) -> Result<PgQueryResult, sqlx::Error> {
    let _ = sqlx::query!("
        WITH names_insert AS(
           INSERT INTO attr_names (name)
           SELECT * FROM UNNEST($1::varchar[])
           ON CONFLICT (name) DO NOTHING
           RETURNING id
        ),
        values_insert AS(
            INSERT INTO attr_vals (value)
            SELECT * FROM UNNEST($2::varchar[])
            ON CONFLICT ON CONSTRAINT attr_vals_uniq DO NOTHING
            RETURNING id
        ),
        names AS(
            SELECT id FROM names_insert
            UNION
            SELECT id FROM attr_vals WHERE (value) IN (SELECT * FROM UNNEST($1::varchar[]))
        ),
        values AS(
            SELECT id FROM values_insert
            UNION
            SELECT id FROM attr_names WHERE (name) IN (SELECT * FROM UNNEST($2::varchar[]))
        )
        INSERT INTO movement_attrs (departure_request, stateless, key, name_id, value_id)
        SELECT * FROM UNNEST($3::uuid[], $4::varchar[], $5::integer[], array(select id from names)::integer[], array(select id from values)::integer[]) as t(departure_request, stateless, key, name_id, value_id)
        WHERE name_id IS NOT NULL AND value_id IS NOT NULL
        ON CONFLICT ON CONSTRAINT mov_attr_pkey DO NOTHING
        ",
        &arrivals
            .iter()
            .map(|a| a.attrs.iter().map(|a| a.value.clone()).collect::<Vec<_>>())
            .flatten()
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|a| a.attrs.iter().map(|a| a.name.clone()).collect::<Vec<_>>())
            .flatten()
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|a| a
                .attrs
                .iter()
                .map(|_| a.departure_request_id.clone())
                .collect::<Vec<_>>())
            .flatten()
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|a| a
                .attrs
                .iter()
                .map(|_| a.stateless.clone())
                .collect::<Vec<_>>())
            .flatten()
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|a| a.attrs.iter().map(|_| a.key).collect::<Vec<_>>())
            .flatten()
            .collect::<Vec<_>>(),
    )
    .fetch_all(conn)
    .await?;

    let operators = sqlx::query_scalar!(
        "WITH result AS(
            INSERT INTO operators (code, name, pub_code)
            SELECT * FROM UNNEST($1::varchar[], $2::varchar[], $3::varchar[])
            ON CONFLICT (code) DO NOTHING
            RETURNING id
         )
         SELECT id FROM result
         UNION
         SELECT id FROM operators WHERE code = ANY($1);",
        &arrivals
            .iter()
            .map(|a| a.operator_code.clone())
            .collect::<Vec<_>>() as &[Option<_>],
        &arrivals
            .iter()
            .map(|a| a.operator_name.clone())
            .collect::<Vec<_>>() as &[Option<_>],
        &arrivals
            .iter()
            .map(|a| a.operator_public_code.clone())
            .collect::<Vec<_>>() as &[Option<_>]
    )
    .fetch_all(conn)
    .await?;

    sqlx::query!(
        "INSERT INTO movements (
        departure_request, stopID, x, y, mapName, area, platform, platformName,
        countdown, time, realtime, key, code,
        number, symbol, motType, mtSubcode, is_realtime, direction, directionFrom,
        name, delay, destID, stateless, lineDisplay, operator_id, type
        )
        SELECT * FROM UNNEST(
            $1::uuid[], $2::integer[], $3::double precision[], $4::double precision[],
            $5::varchar[], $6::varchar[], $7::varchar[], $8::varchar[], $9::integer[],
            $10::timestamp[], $11::timestamp[], $12::integer[], $13::varchar[], $14::varchar[],
            $15::varchar[], $16::integer[], $17::integer[], $18::boolean[],
            $19::varchar[], $20::varchar[], $21::varchar[], $22::integer[],
            $23::varchar[], $24::varchar[], $25::varchar[], $26::integer[], $27::boolean[]
        ) ON CONFLICT ON CONSTRAINT movements_uniq DO NOTHING",
        &arrivals
            .iter()
            .map(|r| r.departure_request_id)
            .collect::<Vec<_>>(),
        &arrivals.iter().map(|r| r.stop_id).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.x).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.y).collect::<Vec<_>>() as &[Option<_>],
        &arrivals
            .iter()
            .map(|r| r.map_name.clone())
            .collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.area.clone()).collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.platform.clone())
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.platform_name.clone())
            .collect::<Vec<_>>(),
        &arrivals.iter().map(|r| r.countdown).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.time).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.real_time).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.key.clone()).collect::<Vec<_>>(),
        &arrivals.iter().map(|r| r.code.clone()).collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.number.clone())
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.symbol.clone())
            .collect::<Vec<_>>(),
        &arrivals.iter().map(|r| r.mot_type).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.mt_subcode).collect::<Vec<_>>() as &[Option<_>],
        &arrivals.iter().map(|r| r.realtime).collect::<Vec<_>>() as &[Option<_>],
        &arrivals
            .iter()
            .map(|r| r.direction.clone())
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.direction_from.clone())
            .collect::<Vec<_>>(),
        &arrivals.iter().map(|r| r.name.clone()).collect::<Vec<_>>(),
        &arrivals.iter().map(|r| r.delay).collect::<Vec<_>>() as &[Option<_>],
        &arrivals
            .iter()
            .map(|r| r.dest_id.clone())
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.stateless.clone())
            .collect::<Vec<_>>(),
        &arrivals
            .iter()
            .map(|r| r.line_display.clone())
            .collect::<Vec<_>>(),
        &operators as &[Option<_>],
        &arrivals.iter().map(|r| r.r#type).collect::<Vec<_>>(),
    )
    .execute(conn)
    .await
}

fn datetime_from_dto(dto: &DateTime2) -> Option<NaiveDateTime> {
    NaiveDate::from_ymd_opt(
        i32::from_str(&dto.year).unwrap(),
        u32::from_str(&dto.month).unwrap(),
        u32::from_str(&dto.day).unwrap(),
    )?
    .and_hms_opt(
        u32::from_str(&dto.hour).unwrap(),
        u32::from_str(&dto.minute).unwrap(),
        0,
    )
}
